
"""Module computing the best coefficient of a QC automatically"""

# Required imports
from pybsrnqc import open_data as od
from pybsrnqc import qc_functions as qcf
from pybsrnqc import plot_limits as pl
from pybsrnqc import coef_study as cs

import importlib.resources
import json


# Coefficients initialisation
# Get data conf from JSON file
with importlib.resources.path("pybsrnqc", "qcrad_conf.json") as data_path:
    with open(data_path, 'r') as f:
        dic_coefs = json.load(f)


def compute(path=None, level='level_2', bw_sel=None):

    # Choosing the QC
    QC = input('What QC do you want to study ? \n Enter QCn where n is the number of the QC chosen \n Answer : ')
    level = input('What level do you want to study ? \n Enter level_1 for 1rst level \n Enter level_2 for 2nd level \n Answer :')

    if QC == 'QC1':
        QC = qcf.QC1()
    if QC == 'QC2':
        QC = qcf.QC2()
    if QC == 'QC3':
        QC = qcf.QC3()
    if QC == 'QC5':
        QC = qcf.QC5()
    if QC == 'QC10':
        QC = qcf.QC10()

    # Get datas
    if path is not None:
        df = od.open_all(path=path + '/')
    else:
        df = od.open_all()

    # Calculating kernel density for the dataset
    if QC.name == 'QC3' or QC.name == 'QC10':
        log_kernel, selected = pl.kde_computing(df, QC, select=True, bw_sel=bw_sel)
    else:
        log_kernel = pl.kde_computing(df, QC, bw_sel=bw_sel)

    # Choosing the threshold
    threshold = float(input('Threshold for outliers : '))

    # Nb of coefficients tried
    nb_try = float(input("Number of coefficients tried (200 is ok): "))

    # Field of research
    coef_range = QC.coef_range
    step = (coef_range[1] - coef_range[0]) / nb_try

    if QC.vary == 'downward_avg':
        coef_range_min = QC.coef_range_min
        step_min = (coef_range_min[1] - coef_range_min[0]) / nb_try

    # Finding the best coefficient for a density threshold given
    if QC.vary == 'downward_avg':
        if QC.name == 'QC10':
            df_score, score, df_score_min, score_min = cs.calc_coef(df, log_kernel, QC, threshold, level=level,
                                                                    coef_range=coef_range, coef_range_min=coef_range_min,
                                                                    step=step, step_min=step_min, selected=selected)

        else:
            df_score, score, df_score_min, score_min = cs.calc_coef(df, log_kernel, QC, threshold, level=level,
                                                                    coef_range=coef_range, coef_range_min=coef_range_min,
                                                                    step=step, step_min=step_min)

    elif QC.vary == 'direct_avg':
        df_score, score = cs.calc_coef(df, log_kernel, QC, threshold, level=level, coef_range=coef_range, step=step, selected=selected)
        print('Best coefficient:')
        print(score)
    else:
        df_score, score = cs.calc_coef(df, log_kernel, QC, threshold, level=level, coef_range=coef_range, step=step)
        print('Best coefficient:')
        print(score)

    # In the case there are several best scores
    if score.shape != (1, 5):
        score = score.iloc[0]
    if QC.vary == 'downward_avg':
        score_min = score_min.iloc[0]

    # Charging the new coeff
    if QC.vary == 'downward_avg':
        dic_coefs['COEF'][QC.coefficients[level]] = float(score[QC.coefficients[level]])
        dic_coefs['COEF'][QC.coefficients[level + '_min']] = float(score_min[QC.coefficients[level + '_min']])
    else:
        dic_coefs['COEF'][QC.coefficients[level]] = float(score[QC.coefficients[level]])

    # Plotting the result
    pl.plot_kde(df, log_kernel, QC, dic_coefs, level=level)

    if QC.vary == 'downward_avg':
        return QC.coefficients[level], float(score[QC.coefficients[level]]), QC.coefficients[level + '_min'], float(score_min[QC.coefficients[level + '_min']])
    else:
        return QC.coefficients[level], float(score[QC.coefficients[level]])



"""
Module that provides the QC class with their assiocated functions and labeling
"""


# Imports required
import math
import numpy as np
from pybsrnqc.config import Coef

# Physics variables


class ref:
    """BSRN Various quantities related to measurements"""
    Tsnw = None  # temperature limit for albedo limit test, temp at which "snow" limit is used
    SOLAR_CONSTANT = 1368  # solar constant at mean Earth-Sun distance uses by QCRad in W.m-2
    BOLTZMANN = 5.67e-8  # Stephan-Boltzman constant = 5.67 x 10 -8 Wm -2 K -4


"""
Flag Value: Related to Type test:
5-6 Global Physical Limits (PP)
3-4 User configurable (UC2) 2nd level tests, also LW Tc and Td tests
1-2 User configurable (UC1) 1st level tests and non-definitive tests
0   No test failures
-1  Missing data or test not possible
"""


''' *********** QC1-QC6 (GSW, Diffuse SW, Direct SW, SWup, LWdn and LWup basic limits tests) *********** '''

# Definition of the QC class


class QC1:

    def __init__(self):
        self.name = 'QC1'
        self.varx = 'SZA'
        self.vary = 'global2_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1': 'C1', 'level_2': 'D1'}
        self.coef_range = [0.0, 1.2]

    def f(self, SZA, GSW, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1 = ref.SOLAR_CONSTANT * coef.C1 * math.pow(math.cos(math.radians(SZA)), 1.2) + 50
        l2 = ref.SOLAR_CONSTANT * coef.D1 * math.pow(math.cos(math.radians(SZA)), 1.2) + 55
        l_bsrn = ref.SOLAR_CONSTANT * 1.5 * math.pow(math.cos(math.radians(SZA)), 1.2) + 100

        return l1, l2, l_bsrn

    def lab(self, SZA, GSW, coef: Coef):
        """GSW [basic limits tests]"""
        if GSW is not None:
            if (GSW < -4):
                return 5
            if (GSW < -2):
                return 3
            if SZA >= 0 and SZA <= 90:  # cannot have float exponent of negative value
                if(GSW > (ref.SOLAR_CONSTANT * 1.5 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 100):
                    return 6
                if(GSW > (ref.SOLAR_CONSTANT * coef.D1 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 55):
                    return 4
                if(GSW > (ref.SOLAR_CONSTANT * coef.C1 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 50):
                    return 2
            else:
                return -1
        else:
            return -1
        return 0

    def calc_lim(self, df, coef: Coef):
        """ Computation of the limits and the labels of a dataframe"""

        df = df[['timestamp', self.varx, self.vary]]

        X_val1 = np.array(df[self.varx])
        X_val2 = np.array(df[self.vary])

        lim_l1 = []
        lim_l2 = []
        lim_bsrn = []

        lim_l1_min = []
        lim_l2_min = []
        lim_bsrn_min = []

        labels = []

        for VAL1, VAL2 in zip(X_val1, X_val2):

            if self.vary == 'downward_avg':

                l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = self.f(VAL1, VAL2, coef)

                lim_l1_min.append(l1_min)
                lim_l2_min.append(l2_min)
                lim_bsrn_min.append(l_bsrn_min)

            else:

                l1, l2, l_bsrn = self.f(VAL1, VAL2, coef)

            label = self.lab(VAL1, VAL2, coef)

            lim_l1.append(l1)
            lim_l2.append(l2)
            lim_bsrn.append(l_bsrn)
            labels.append(label)

        return(lim_l1, lim_l2, lim_bsrn, labels)


class QC2:

    def __init__(self):
        self.name = 'QC2'
        self.varx = 'SZA'
        self.vary = 'diffuse_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1': 'C2', 'level_2': 'D2'}
        self.coef_range = [0.0, 1.0]

    def f(self, SZA, Dif, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1 = ref.SOLAR_CONSTANT * coef.C2 * math.pow(math.cos(math.radians(SZA)), 1.2) + 30
        l2 = ref.SOLAR_CONSTANT * coef.D2 * math.pow(math.cos(math.radians(SZA)), 1.2) + 35
        l_bsrn = ref.SOLAR_CONSTANT * 0.95 * math.pow(math.cos(math.radians(SZA)), 1.2) + 50

        return l1, l2, l_bsrn

    def lab(self, SZA, Dif, coef: Coef):
        """Diffuse SW [basic limits tests]"""
        if Dif is not None:
            if (Dif < -4):
                return 5
            if(Dif < -2):
                return 3
            if SZA >= 0 and SZA <= 90:  # cannot have float exponent of negative value
                if(Dif > (ref.SOLAR_CONSTANT * 0.95 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 50):
                    return 6
                if(Dif > (ref.SOLAR_CONSTANT * coef.D2 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 35):
                    return 4
                if(Dif > (ref.SOLAR_CONSTANT * coef.C2 * math.pow(math.cos(math.radians(SZA)), 1.2)) + 30):
                    return 2
            else:
                return -1
        else:
            return -1
        return 0

    def calc_lim(self, df, coef: Coef):
        """ Computation of the limits and the labels of a dataframe"""

        df = df[['timestamp', self.varx, self.vary]]

        X_val1 = np.array(df[self.varx])
        X_val2 = np.array(df[self.vary])

        lim_l1 = []
        lim_l2 = []
        lim_bsrn = []

        lim_l1_min = []
        lim_l2_min = []
        lim_bsrn_min = []

        labels = []

        for VAL1, VAL2 in zip(X_val1, X_val2):

            if self.vary == 'downward_avg':

                l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = self.f(VAL1, VAL2, coef)

                lim_l1_min.append(l1_min)
                lim_l2_min.append(l2_min)
                lim_bsrn_min.append(l_bsrn_min)

            else:

                l1, l2, l_bsrn = self.f(VAL1, VAL2, coef)

            label = self.lab(VAL1, VAL2, coef)

            lim_l1.append(l1)
            lim_l2.append(l2)
            lim_bsrn.append(l_bsrn)
            labels.append(label)

        return(lim_l1, lim_l2, lim_bsrn, labels)


class QC3:

    def __init__(self):
        self.name = 'QC3'
        self.varx = 'SZA'
        self.vary = 'direct_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1': 'C3', 'level_2': 'D3'}
        self.coef_range = [0.0, 1.0]

    def f(self, SZA, DirN, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1 = ref.SOLAR_CONSTANT * coef.C3 * math.pow(math.cos(math.radians(SZA)), 0.2) + 10
        l2 = ref.SOLAR_CONSTANT * coef.D3 * math.pow(math.cos(math.radians(SZA)), 0.2) + 15
        l_bsrn = ref.SOLAR_CONSTANT

        return l1, l2, l_bsrn

    def lab(self, SZA, DirN, coef: Coef):
        """Direct Normal SW  [basic limits tests]"""
        if DirN is not None:
            if (DirN < -4):
                return 5
            if(DirN < -2):
                return 3
            if (DirN > ref.SOLAR_CONSTANT):
                return 6
            if SZA >= 0 and SZA <= 90:  # cannot have float exponent of negative value
                if(DirN > (ref.SOLAR_CONSTANT * coef.D3 * math.pow(math.cos(math.radians(SZA)), 0.2)) + 15):
                    return 4
                if(DirN > (ref.SOLAR_CONSTANT * coef.C3 * math.pow(math.cos(math.radians(SZA)), 0.2)) + 10):
                    return 2
            else:
                return -1
        else:
            return -1
        return 0

    def calc_lim(self, df, coef: Coef):
        """ Computation of the limits and the labels of a dataframe"""

        df = df[['timestamp', self.varx, self.vary]]

        X_val1 = np.array(df[self.varx])
        X_val2 = np.array(df[self.vary])

        lim_l1 = []
        lim_l2 = []
        lim_bsrn = []

        lim_l1_min = []
        lim_l2_min = []
        lim_bsrn_min = []

        labels = []

        for VAL1, VAL2 in zip(X_val1, X_val2):

            if self.vary == 'downward_avg':

                l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = self.f(VAL1, VAL2, coef)

                lim_l1_min.append(l1_min)
                lim_l2_min.append(l2_min)
                lim_bsrn_min.append(l_bsrn_min)

            else:

                l1, l2, l_bsrn = self.f(VAL1, VAL2, coef)

            label = self.lab(VAL1, VAL2, coef)

            lim_l1.append(l1)
            lim_l2.append(l2)
            lim_bsrn.append(l_bsrn)
            labels.append(label)

        return(lim_l1, lim_l2, lim_bsrn, labels)


class QC5:

    def __init__(self):
        self.name = 'QC5'
        self.varx = 'SZA'
        self.vary = 'downward_avg'
        self.unitx = 'SZA'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1_min': 'C5', 'level_2_min': 'D5', 'level_1': 'C6', 'level_2': 'D6'}
        self.coef_range = [400.0, 600.0]
        self.coef_range_min = [200.0, 400.0]

    def f(self, SZA, LWdn, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1_max = coef.C6
        l2_max = coef.D6
        l_bsrn_max = 700

        l1_min = coef.C5
        l2_min = coef.D5
        l_bsrn_min = 40

        return l1_max, l2_max, l_bsrn_max, l1_min, l2_min, l_bsrn_min

    def lab(self, Ta, LWdn, coef: Coef):
        """LWdn [basic limits tests]"""
        if LWdn is not None:
            if (LWdn > 700):
                return 6
            if (LWdn < 40):
                return 5
            if (LWdn > coef.D6):
                return 4
            if (LWdn < coef.D5):
                return 3
            if (LWdn > coef.C6):
                return 2
            if (LWdn < coef.C5):
                return 1
        else:
            return -1
        return 0

    def calc_lim(self, df, coef: Coef):
        """ Computation of the limits and the labels of a dataframe"""

        df = df[['timestamp', self.varx, self.vary]]

        X_val1 = np.array(df[self.varx])
        X_val2 = np.array(df[self.vary])

        lim_l1 = []
        lim_l2 = []
        lim_bsrn = []

        lim_l1_min = []
        lim_l2_min = []
        lim_bsrn_min = []

        labels = []

        for VAL1, VAL2 in zip(X_val1, X_val2):

            l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = self.f(VAL1, VAL2, coef)

            lim_l1_min.append(l1_min)
            lim_l2_min.append(l2_min)
            lim_bsrn_min.append(l_bsrn_min)

            label = self.lab(VAL1, VAL2, coef)

            lim_l1.append(l1)
            lim_l2.append(l2)
            lim_bsrn.append(l_bsrn)
            labels.append(label)

        return(lim_l1, lim_l2, lim_bsrn, labels)


class QC10:

    def __init__(self):
        self.name = 'QC10'
        self.varx = 'temperature'
        self.vary = 'downward_avg'
        self.unitx = '°K'
        self.unity = 'W.m-2'
        self.coefficients = {'level_1_min': 'C11', 'level_2_min': 'D11', 'level_1': 'C12', 'level_2': 'D12'}
        self.coef_range = [0.0, 10.0]
        self.coef_range_min = [0.60, 1.0]

    def f(self, Ta, LWdn, coef: Coef):
        ''' Return the 2 main variables, the 1rst level limit, the 2nd level limit and the physical limit for a sample'''

        l1_max = ref.BOLTZMANN * math.pow(Ta + 273.15, 4) + coef.C12
        l2_max = ref.BOLTZMANN * math.pow(Ta + 273.15, 4) + coef.D12
        l_bsrn_max = 700

        l1_min = coef.C11 * ref.BOLTZMANN * math.pow(Ta + 274.15, 4)
        l2_min = coef.D11 * ref.BOLTZMANN * math.pow(Ta + 274.15, 4)
        l_bsrn_min = 40

        return l1_max, l2_max, l_bsrn_max, l1_min, l2_min, l_bsrn_min

    def lab(self, Ta, LWdn, coef: Coef):
        """LWdn to Ta test"""
        if all(v is not None for v in [LWdn, Ta]):
            if QC19(Ta) == 0:
                if (LWdn > (ref.BOLTZMANN * math.pow(Ta, 4) + coef.D12)):
                    return 4
                if (LWdn < (coef.D11 * ref.BOLTZMANN * math.pow(Ta, 4))):
                    return 3
                if (LWdn > (ref.BOLTZMANN * math.pow(Ta, 4) + coef.C12)):
                    return 2
                if (LWdn < (coef.C11 * ref.BOLTZMANN * math.pow(Ta, 4))):
                    return 1
            else:
                return -1
        else:
            return -1
        return 0

    def calc_lim(self, df, coef: Coef):
        """ Computation of the limits and the labels of a dataframe"""

        df = df[['timestamp', self.varx, self.vary]]

        X_val1 = np.array(df[self.varx]) + 273.15
        X_val2 = np.array(df[self.vary])

        lim_l1 = []
        lim_l2 = []
        lim_bsrn = []

        lim_l1_min = []
        lim_l2_min = []
        lim_bsrn_min = []

        labels = []

        for VAL1, VAL2 in zip(X_val1, X_val2):

            l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = self.f(VAL1, VAL2, coef)

            lim_l1_min.append(l1_min)
            lim_l2_min.append(l2_min)
            lim_bsrn_min.append(l_bsrn_min)

            label = self.lab(VAL1, VAL2, coef)

            lim_l1.append(l1)
            lim_l2.append(l2)
            lim_bsrn.append(l_bsrn)
            labels.append(label)

        return(lim_l1, lim_l2, lim_bsrn, labels)

# -------------------------------------------------------------------------------------------


def QC19(Ta):
    """Ta testing"""
    if Ta is not None:
        if Ta > 350 or Ta < 170:
            return 1
    else:
        return -1
    return 0



#!/usr/bin/env python3
import csv
import os
import pandas as pd
import datetime
from collections import OrderedDict
from pathlib import Path
from bokeh.layouts import column
from bokeh.models import ColumnDataSource
from bokeh.models.tools import HoverTool
from bokeh.plotting import figure, output_file, show
from pybsrnqc import utils
from pybsrnqc.config import Station, Coef, Header
from pybsrnqc import qc_functions as qcf


default_station = Station()
default_coef = Coef()
default_header = Header()


def fix_values(row: OrderedDict, header: Header = default_header):
    GSW, Dif, DirN, LWdn, Ta, Td = None, None, None, None, None, None
    # check variables name on input file
    try:
        timestamp = row[header.TIMESTAMP_NAME]
        if header.GSW_NAME in row and utils.isfloat(row[header.GSW_NAME]):
            GSW = float(row[header.GSW_NAME])
        if header.DIF_NAME in row and utils.isfloat(row[header.DIF_NAME]):
            Dif = float(row[header.DIF_NAME])
        if header.DIR_NAME in row and utils.isfloat(row[header.DIR_NAME]):
            DirN = float(row[header.DIR_NAME])
        if header.LWDN_NAME in row and utils.isfloat(row[header.LWDN_NAME]):
            LWdn = float(row[header.LWDN_NAME])
        if header.TA_NAME in row and utils.isfloat(row[header.TA_NAME]):
            Ta = float(row[header.TA_NAME]) + 273.15
    except KeyError as e:
        print('KeyError: %s' % str(e))
    except ValueError as valueError:
        print('ValueError: %s' % str(valueError))
    return (timestamp, GSW, Dif, DirN, LWdn, Ta)


def getRow(row: OrderedDict, zenith_serie, coef: Coef = default_coef, header: Header = default_header):
    # get values of parameters
    [timestamp, GSW, Dif, DirN, LWdn, Ta] = fix_values(row, header)
    SZA = float(zenith_serie[timestamp])
    # application of data quality control
    qc_result = {
        "timestamp": timestamp,
        "QC1": qcf.QC1().lab(SZA, GSW, coef),
        "QC2": qcf.QC2().lab(SZA, Dif, coef),
        "QC3": qcf.QC3().lab(SZA, DirN, coef),
        "QC5": qcf.QC5().lab(SZA, LWdn, coef),
        "QC10": qcf.QC10().lab(Ta, LWdn, coef),
        "QC19": qcf.QC19(Ta)
    }
    # create row for aqc file
    row_aqc = row
    for key in row_aqc.keys():
        if key != "timestamp":
            row_aqc[key] = 0
    if qc_result["QC1"] > 2:
        row_aqc["global2_avg"] = 1
        row_aqc["global2_std"] = 1
        row_aqc["global2_min"] = 1
        row_aqc["global2_max"] = 1
    if qc_result["QC2"] > 2:
        row_aqc["diffuse_avg"] = 1
        row_aqc["diffuse_std"] = 1
        row_aqc["diffuse_min"] = 1
        row_aqc["diffuse_max"] = 1
    if qc_result["QC3"] > 2:
        row_aqc["direct_avg"] = 1
        row_aqc["direct_std"] = 1
        row_aqc["direct_min"] = 1
        row_aqc["direct_max"] = 1
    if qc_result["QC5"] > 2:
        row_aqc["downward_avg"] = 1
        row_aqc["downward_std"] = 1
        row_aqc["downward_min"] = 1
        row_aqc["downward_max"] = 1
    if qc_result["QC19"] > 0:
        row_aqc["temperature"] = 1
    # create row for qcrad file
    row_qcrad = qc_result
    return row_aqc, row_qcrad


def generateQCFiles(filepath, station: Station = default_station, coef: Coef = default_coef):
    """Create the 2 files _aqc.csv and _qrcad.csv"""
    print(filepath)
    # Manage filenames
    FILE_BRUT = filepath
    FILE_AQC = os.path.splitext(FILE_BRUT.replace('_brut', ''))[0] + '_aqc.csv'
    FILE_QCRAD = os.path.splitext(FILE_BRUT.replace('_brut', ''))[0] + '_qcrad.csv'
    # load input file into a DataFrame
    file_brut = pd.read_csv(FILE_BRUT)
    timestamp_list = file_brut.timestamp.to_list()
    zenith_serie = utils.getZenith(timestamp_list, station.LAT, station.LON, station.ALT)
    # process input file
    datapoints_qcrad, datapoints_aqc = [], []
    with open(FILE_BRUT, 'r') as fileIn:
        reader = csv.DictReader(fileIn, delimiter=',')
        for row in reader:
            (row_aqc, row_qcrad) = getRow(row, zenith_serie, coef)
            datapoints_qcrad.append(row_qcrad)
            datapoints_aqc.append(row_aqc)
    # convert list of qcrad results into dataframe & save as a csv file
    df_qcrad = pd.DataFrame(datapoints_qcrad)
    df_qcrad.to_csv(FILE_QCRAD, index=False)
    print("Successfully saved CSV file in", FILE_QCRAD)
    # convert list of aqc results into dataframe & save as a csv file
    df_aqc = pd.DataFrame(datapoints_aqc)
    df_aqc.to_csv(FILE_AQC, index=False)
    print("Successfully saved CSV file in", FILE_AQC)


def plotQCFiles(filepath):
    """Plot the input file"""
    # init
    csv_name = filepath
    df = pd.read_csv(filepath)
    year = df['timestamp'][0][:4]
    month = df['timestamp'][0][5:7]
    yearMonth = year + month
    typeFile = 'brut'
    # create directory if it necessary
    Path("plot/" + yearMonth).mkdir(parents=True, exist_ok=True)
    chart_name = 'plot/' + yearMonth + '/' + typeFile + '.html'
    if typeFile == "qcrad":
        parameters_to_plot = ["QC1", "QC2", "QC3", "QC5", "QC7", "QC8", "QC10", "QC19"]
    else:
        parameters_to_plot = ["global2_avg", "diffuse_avg", "direct_avg", "downward_avg", "temperature"]
    # get sensor data
    df = pd.read_csv(csv_name)
    timestamp = df.timestamp.to_list()
    date = []
    for time in timestamp:
        date.append(datetime.datetime.strptime(time, '%Y-%m-%d %H:%M:%S'))
    df.insert(1, "date", date)
    source = ColumnDataSource(df)
    # create plots and choose color
    plots = []
    for var in parameters_to_plot:
        hover = HoverTool()
        s = figure(plot_width=1600, plot_height=900, x_axis_type="datetime")
        s.circle(x='date', y=var, source=source)
        s.line(x='date', y=var, legend_label=var, source=source)
        hover.tooltips = [
            ('date', '@date{%F %T}'),
            (var, '@' + var)
        ]
        hover.formatters = {'@date': 'datetime'}
        s.add_tools(hover)
        plots.append(s)
    # show and save plot
    output_file(chart_name)
    print("Successfully saved HTML file in", chart_name)
    show(column(plots))



"""
Module that plot the curves following the quality controle equations
 """

# Required imports

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pickle
import os

from scipy import stats

from matplotlib.widgets import LassoSelector
from matplotlib.path import Path

# -----------------------------------------------------------------------------------------------------------
# Plots of the BSRN limits on datasets


def limit_plot(df, QC, coefs, save: bool = False, level='all', display: bool = True,
               values: bool = True, fig: bool = True):

    """ Function plotting the limit curves and the dataset points of a
        dataframe"""

    df = df[['timestamp', QC.varx, QC.vary]]

    X_val1 = np.array(df[QC.varx])
    X_val2 = np.array(df[QC.vary])

    lim_l1 = []
    lim_l2 = []
    lim_bsrn = []

    lim_l1_min = []
    lim_l2_min = []
    lim_bsrn_min = []

    val1_var = list(X_val1)
    val2_var = list(X_val2)

    for VAL1, VAL2 in zip(X_val1, X_val2):

        if QC.vary == 'downward_avg':

            l1, l2, l_bsrn, l1_min, l2_min, l_bsrn_min = QC.f(VAL1, VAL2,
                                                              coefs)

            lim_l1_min.append(l1_min)
            lim_l2_min.append(l2_min)
            lim_bsrn_min.append(l_bsrn_min)

        else:

            l1, l2, l_bsrn = QC.f(VAL1, VAL2, coefs)

        lim_l1.append(l1)
        lim_l2.append(l2)
        lim_bsrn.append(l_bsrn)

    if fig:
        plt.figure(figsize=(20, 14))

    if values:
        plt.scatter(val1_var, val2_var, marker='+', color='dodgerblue',
                    label='Values', s=1)

    if level == 'all':
        plt.scatter(val1_var, lim_l1, marker='+', color='green',
                    label=f"1rst level limit - {coefs['COEF'][QC.coefficients['level_1']]}", s=1)
        plt.scatter(val1_var, lim_l2, marker='+', color='blue',
                    label=f"2nd level limit - {coefs['COEF'][QC.coefficients['level_2']]}", s=1)
        plt.scatter(val1_var, lim_bsrn, marker='+', color='red',
                    label="BSRN  limit", s=1)
    if level == 'level_1':
        plt.scatter(val1_var, lim_l1, marker='+', color='green',
                    label=f"1rst level limit - {coefs['COEF'][QC.coefficients['level_1']]}", s=1)
    if level == 'level_2':
        plt.scatter(val1_var, lim_l2, marker='+', color='blue',
                    label=f"2nd level limit - {coefs['COEF'][QC.coefficients['level_2']]}", s=1)
    if level == 'level_bsrn':
        plt.scatter(val1_var, lim_bsrn, marker='+', color='red', label="BSRN  limit", s=1)

    if QC.vary == 'downward_avg':

        if level == 'all':
            plt.scatter(val1_var, lim_l1_min, marker='+', color='lightgreen', label=f"1rst level limit - {coefs['COEF'][QC.coefficients['level_1_min']]}", s=1)
            plt.scatter(val1_var, lim_l2_min, marker='+', color='lightblue', label=f"2nd level limit - {coefs['COEF'][QC.coefficients['level_2_min']]}", s=1)
            plt.scatter(val1_var, lim_bsrn_min, marker='+', color='lightcoral', label="BSRN  limit", s=1)
        if level == 'level_1':
            plt.scatter(val1_var, lim_l1_min, marker='+', color='lightgreen', label=f"1rst level limit - {coefs['COEF'][QC.coefficients['level_1_min']]}", s=1)
        if level == 'level_2':
            plt.scatter(val1_var, lim_l2_min, marker='+', color='lightblue', label=f"2nd level limit - {coefs['COEF'][QC.coefficients['level_2_min']]}", s=1)
        if level == 'level_bsrn':
            plt.scatter(val1_var, lim_bsrn_min, marker='+', color='lightcoral', label="BSRN  limit", s=1)

    if save:
        plt.savefig(f'limit_plot - {QC.vary} versus {QC.varx}')
    if display:
        plt.xlabel(QC.unitx,)
        plt.ylabel(QC.unity)
        plt.title(f'Limits - {QC.vary}', fontsize=20)
        plt.legend(prop={'size': 15})
        plt.show()


def multiplot_coef(df, QC, coefs, level='level_2', coef_values=[0.0, 0.5, 1.2], level_min=None, coef_values_min=None):
    'Dessine la courbe limite pour plusieurs valeurs du coefficients'

    X_val1 = np.array(df[QC.varx])
    X_val2 = np.array(df[QC.vary])

    val1_var = list(X_val1)
    val2_var = list(X_val2)

    plt.figure(figsize=(20, 14))
    plt.scatter(val1_var, val2_var, marker='+', color='dodgerblue',
                label='Values', s=1)

    for v in coef_values:

        # We compute the considered coefficient

        new_coef = coefs.copy()
        new_coef['COEF'][QC.coefficients[level]] = v

        # We plot for this coefficient

        limit_plot(df, QC, new_coef, level=level, display=False,
                   values=False, fig=False)

    if coef_values_min is not None:
        for v in coef_values_min:

            # We compute the considered coefficient

            new_coef = coefs.copy()
            new_coef['COEF'][QC.coefficients[level_min]] = v

            # We plot for this coefficient

            limit_plot(df, QC, new_coef, level=level, display=False,
                       values=False, fig=False)

    plt.xlabel(QC.unitx)
    plt.ylabel(QC.unity)
    plt.title(f'Limits - {QC.vary}', fontsize=20)
    plt.legend(prop={'size': 15})
    plt.show()


# ----------------------------------------------------------------------------------------------------------------------
# Displaying of the data histograms in 2D and 3D. It allows us to see the
# shape of the data and its caracteristics like unimodality

def hist_data(df, QC, dimension='3D'):

    # Get the data
    X = np.array(df[[QC.vary, QC.varx]])

    # Create the figure
    fig = plt.figure(figsize=(20, 14))
    ax = fig.add_subplot(projection='3d')

    hist, xedges, yedges = np.histogram2d(X[:, 0], X[:, 1], bins=75)

    # Construct arrays for the anchor positions of the 16 bars.
    xpos, ypos = np.meshgrid(xedges[:-1] + 0.25, yedges[:-1] + 0.25, indexing="ij")
    xpos = xpos.ravel()
    ypos = ypos.ravel()
    zpos = 0

    # Construct arrays with the dimensions for the 16 bars.
    dx = xedges[-1] / 75 * np.ones_like(zpos)
    dy = yedges[-1] / 75 * np.ones_like(zpos)
    dz = hist.ravel()

    cmap = plt.cm.get_cmap('Wistia')  # Get desired colormap - you can change this
    max_height = np.max(dz)   # get range of colorbars so we can normalize
    min_height = np.min(dz)
    # scale each z to [0,1], and get their rgb values
    rgba = [cmap((k - min_height) / max_height) for k in dz]

    ax.bar3d(xpos, ypos, zpos, dx, dy, dz, color=rgba, zsort='average')

    plt.title(f'Data histogram - {QC.vary} vs {QC.varx}', fontdict={'fontsize': 20})
    plt.xlabel(QC.unity)
    plt.ylabel(QC.unitx)
    ax.set_zlabel('Number of samples')
    plt.show()


# --------------------------------------------------------------------------------------------
# KDE computation and plotting for our dataset


def kde_computing(df, QC, display=True, coefs=None, limits=False, level='All',
                  log_form=True, save='KDE_result', select=False, bw_sel=None):

    # Get the data
    X = np.array(df[[QC.vary, QC.varx]]).T
    # Kernel calculation
    print('Computing kde - It can take some times')

    if bw_sel is None:
        kernel = stats.gaussian_kde(X)(X)
        kernel_log = np.log(kernel)
    else:
        kernel = stats.gaussian_kde(X, bw_method=bw_sel)(X)
        kernel_log = np.log(kernel)

    # Plot
    if display:

        fig, ax = plt.subplots(figsize=(20, 14))
        pts = plt.scatter(X[1, :], X[0, :], c=kernel_log,
                          s=1, cmap=plt.cm.jet)
        selector = SelectFromCollection(ax, pts)

    if limits:
        limit_plot(df, QC, coefs, save=False, level='all', display=False,
                   values=False, fig=False)

    if display:
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':

                def accept(event):
                    if event.key == "enter":
                        # print("Selected points:")
                        selected = selector.xys[selector.ind].tolist()
                        with open("selected.txt", "wb") as fp:   # Pickling
                            pickle.dump(selected, fp)
                        # print(selected)
                        # print(type(selected))
                        selector.disconnect()
                        ax.set_title("", fontsize=20)
                        fig.canvas.draw()

                fig.canvas.mpl_connect("key_press_event", accept)
            ax.set_title("Press enter to accept selected points.", fontsize=20)

        ax.set_xlabel(QC.unitx)
        ax.set_ylabel(QC.unity)
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':
                ax.set_title(f"{QC.vary} density along {QC.varx} \n Select the points that you don't want to consider", fontsize=20)
        else:
            ax.set_title(f'{QC.vary} density along {QC.varx}', fontsize=20)
        plt.colorbar(pts, label='Density Log(KDE)')
        ax.legend(prop={'size': 15})
        plt.show()

    # Download file
    if save is not None:
        pd.DataFrame(kernel).to_csv(save + '.csv')

    # Return density
    if log_form:
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':
                with open("selected.txt", "rb") as fp:   # Unpickling
                    selected = pickle.load(fp)
                os.remove("selected.txt")
                return kernel_log, selected
        else:
            return kernel_log
    else:
        if select:
            if QC.name == 'QC3' or QC.name == 'QC10':
                with open("selected.txt", "rb") as fp:   # Unpickling
                    selected = pickle.load(fp)
                os.remove("selected.txt")
                return kernel, selected
        else:
            return kernel


def plot_kde(df, log_kernel, QC, coefs, level='level_2'):

    plt.figure(figsize=(20, 14))

    # Get the data
    X = np.array(df[[QC.vary, QC.varx]]).T

    # Plot the data
    plot = plt.scatter(X[1, :], X[0, :], c=log_kernel,
                       s=1, cmap=plt.cm.jet)

    if level is not None:
        limit_plot(df, QC, coefs, save=False, level=level, display=False,
                   values=False, fig=False)

    plt.xlabel(QC.unitx)
    plt.ylabel(QC.unity)
    plt.title(f'{QC.vary} density along {QC.varx}', fontsize=20)
    plt.colorbar(plot, label='Density Log(KDE)')
    plt.legend(prop={'size': 15})
    plt.show()


def plot_series_kde(df, log_kernel, QC, begin, end,
                    var='timestamp', line=True):

    df['log_kde'] = log_kernel
    kernel_range = [np.min(log_kernel), np.max(log_kernel)]
    mask = (df['timestamp'] >= begin) & (df['timestamp'] < end)
    df_week = df.loc[mask]

    if var == 'SZA':
        tick = None
    else:
        tick = np.arange(0, len(list(df_week[var])), 60)

    plt.figure(figsize=(20, 14))

    if line:
        plt.plot(list(df_week[var]), list(df_week[QC.vary]),
                 color='black', alpha=0.2)

    plot = plt.scatter(list(df_week[var]), list(df_week[QC.vary]),
                       c=np.array(df_week['log_kde']), cmap=plt.cm.jet, s=5)

    plt.xlabel(var)
    plt.ylabel(QC.unity)
    plt.xticks(tick, rotation=45)
    plt.title(f'{QC.vary} and its density - {begin}', fontsize=20)
    plt.clim(kernel_range[0], kernel_range[1])
    plt.colorbar(plot, label='Density Log(KDE)')
    plt.show()


# -----------------------------------------------
# Lasso used to select the datas

class SelectFromCollection:
    """
    Select indices from a matplotlib collection using `LassoSelector`.

    Selected indices are saved in the `ind` attribute. This tool fades out the
    points that are not part of the selection (i.e., reduces their alpha
    values). If your collection has alpha < 1, this tool will permanently
    alter the alpha values.

    Note that this tool selects collection objects based on their *origins*
    (i.e., `offsets`).

    Parameters
    ----------
    ax : `~matplotlib.axes.Axes`
        Axes to interact with.
    collection : `matplotlib.collections.Collection` subclass
        Collection you want to select from.
    alpha_other : 0 <= float <= 1
        To highlight a selection, this tool sets all selected points to an
        alpha value of 1 and non-selected points to *alpha_other*.
    """

    def __init__(self, ax, collection, alpha_other=0.3):
        self.canvas = ax.figure.canvas
        self.collection = collection
        self.alpha_other = alpha_other

        self.xys = collection.get_offsets()
        self.Npts = len(self.xys)

        # Ensure that we have separate colors for each object
        self.fc = collection.get_facecolors()
        if len(self.fc) == 0:
            raise ValueError('Collection must have a facecolor')
        elif len(self.fc) == 1:
            self.fc = np.tile(self.fc, (self.Npts, 1))

        self.lasso = LassoSelector(ax, onselect=self.onselect)
        self.ind = []

    def onselect(self, verts):
        path = Path(verts)
        self.ind = np.nonzero(path.contains_points(self.xys))[0]
        self.fc[:, -1] = self.alpha_other
        self.fc[self.ind, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()

    def disconnect(self):
        self.lasso.disconnect_events()
        self.fc[:, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()
